__author__ = 'Robert Ancell <bob27@users.sourceforge.net>'
__license__ = 'GNU General Public License Version 2'
__copyright__ = 'Copyright 2005-2006  Robert Ancell'

import gtk
import gtk.glade
import gtk.gdk

# TEMP, shared between too many files
import os
UI_DATA_DIR = 'lib/glchess/gtkui/'
if not os.path.isdir(UI_DATA_DIR):
    UI_DATA_DIR = '/usr/share/games/glchess/gui/'
    
class RemoteGame:
    """
    """
    
    __host = ''
    __serverId = None
    __name = ''
    
    def __init__(self, host, serverId, name):
        self.__host = host
        self.__serverId = serverId
        self.__name = name

    def getHost(self):
        return self.__host
    
    def getServerId(self):
        return self.__serverId
    
    def getGameName(self):
        return self.__name
    
class GtkServerList:
    __gui = None
    
    __servers = None
    
    def __init__(self, gui):
        self.__gui = gui
        
        self.__servers = []
        view = gui.get_widget('ng_server_list')
        if view is not None:
            store = gtk.ListStore(str, gobject.TYPE_PYOBJECT)
            view.set_model(store)

            cell = gtk.CellRendererText()
            column = gtk.TreeViewColumn('name', cell)
            column.add_attribute(cell, 'text', 0)
            view.append_column(column)
        
    def add(self, host, serverId, name):
        """
        """
        server = RemoteGame(host, serverId, name)
        self.__servers.append(server)
        
        view = self.__gui.get_widget('ng_server_list')
        if view is None:
            return
        model = view.get_model()
        iter = model.append()
        model.set(iter, 0, name)
        model.set(iter, 1, server)
        
    def getSelected(self):
        """
        """
        view = self.__gui.get_widget('ng_server_list')
        if view is None:
            return None
        selection = view.get_selection()
        (model, iter) = selection.get_selected()
        
        if iter is None:
            return None
        else:
            return model.get_value(iter, 1)
        
    def find(self, host, serverId):
        for s in self.__servers:
            if s.getHost() == host and s.getServerId() == serverId:
                return s
        return None
        
    def remove(self, host, serverId):
        """
        """
        s = self.find(host, serverId)
        try:
            self.__servers.remove(s)
        except:
            pass
        
        view = self.__gui.get_widget('ng_server_list')
        if view is None:
            return
        model = view.get_model()
        
        iter = model.get_iter_first()
        while iter is not None:
            if model.get_value(iter, 1) is s:
                break
            iter = model.iter_next(iter)
        
        if iter is not None:
            model.remove(iter)

class GtkNewGameDialog:
    """
    """
    
    # The main UI and the ???
    __mainUI = None
    __gui = None
    
    __serverList = None
    
    __moves = None
    
    def __init__(self, mainUI, aiModel, gameName = None,
                 whiteName = None, blackName = None,
                 whiteAI = None, blackAI = None, moves = None):
        """Constructor for a new game dialog.
        
        'mainUI' is the main UI.
        'aiModel' is the AI models to use.
        'gameName' is the name of the game (string) or None if unknown.
        'whiteName' is the name of the white player (string) or None if unknown.
        'blackName' is the name of the white player (string) or None if unknown.
        'whiteAI' is the type of AI the white player is (string) or None if no AI.
        'blackAI' is the type of AI the black player is (string) or None if no AI.
        'moves' is a list of moves (strings) that the have already been made.
        """
        self.__mainUI = mainUI
        self.__moves = moves
        
        # Load the UI
        self.__gui = gtk.glade.XML(UI_DATA_DIR + 'new_game_simple.glade', domain = 'glchess')
        self.__gui.signal_autoconnect(self)
        
        # Make all the AI combo boxes use one list of AI types
        widgets = (self.__getWidget('ng_black_ai_type_combo'),
                   self.__getWidget('ng_white_ai_type_combo'),
                   self.__getWidget('ng_local_ai_type_combo'))
        for widget in widgets:
            if widget is None:
                continue
            
            widget.set_model(aiModel)
            cell = gtk.CellRendererText()
            widget.pack_start(cell, False)
            widget.add_attribute(cell, 'text', 0)
            widget.set_active(0)
            
        haveAIs = aiModel.iter_n_children(None) != 0
        self.__getWidget('white_ai_vbox').set_sensitive(haveAIs)
        self.__getWidget('black_ai_vbox').set_sensitive(haveAIs)
        
        # Use the supplied properties
        if moves:
            self.__getWidget('new_game_dialog').set_title('Restore game (%i moves)' % len(moves))
        if gameName:
            self.__getWidget('ng_game_name_entry').set_text(gameName)
            
        if whiteName:
            self.__getWidget('ng_white_name_entry').set_text(whiteName)
        if blackName:
            self.__getWidget('ng_black_name_entry').set_text(blackName)
            
        # Configure AIs
        if whiteAI:
            self.__getWidget('ng_white_ai_type_combo').set_active_iter(self.__getAIIter(aiModel, whiteAI))
            self.__getWidget('ng_white_ai_radio').set_active(True)
        else:
            self.__getWidget('ng_white_human_radio').set_active(True)
        if blackAI:
            self.__getWidget('ng_black_ai_type_combo').set_active_iter(self.__getAIIter(aiModel, blackAI))
            self.__getWidget('ng_black_ai_radio').set_active(True)
        else:
            self.__getWidget('ng_black_human_radio').set_active(True)
        
        # Show the dialog
        self.__getWidget('new_game_dialog').show()
        self.__testReady()
        
        # ...
        self.__serverList = GtkServerList(self.__gui)
        view = self.__getWidget('ng_server_list')
        if view is not None:
            selection = view.get_selection()
            selection.connect('changed', self.on_ng_properties_changed)

        # TEMP
        self.__serverList.add('alchemy', 1, "Bob's server")

    # Private methods
    
    def __getAIIter(self, model, name):
        """Get an AI engine.
        
        'name' is the name of the AI engine to find.
        
        Return the iter for this AI or None if no AI of this name.
        """
        # FIXME: I'm sure there is a more efficient way of doing this...
        iter = model.get_iter_first()
        while True:
            if name == model.get_value(iter, 0):
                return iter
            
            iter = model.iter_next(iter)
            if iter is None:
                return None

    def __getWidget(self, name):
        """
        """
        return self.__gui.get_widget(name)
    
    def __getAIType(self, comboBox):
        """
        """
        model = comboBox.get_model()
        iter = comboBox.get_active_iter()
        if iter is None:
            return None
        
        data = model.get(iter, 0)
        return data[0]
    
    def __getWhitePlayer(self):
        """
        """
        name = self.__getWidget('ng_white_name_entry').get_text()
        if self.__getWidget('ng_white_human_radio').get_active():
            if len(name) == 0:
                return None
            return (name, None)
        
        elif self.__getWidget('ng_white_ai_radio').get_active():
            if len(name) == 0:
                return None
            aiType = self.__getAIType(self.__getWidget('ng_white_ai_type_combo'))
            if aiType is None:
                return None
            return (name, aiType)
        
        elif self.__getWidget('ng_white_network_radio').get_active():
            return (None, None)
        
    def __getBlackPlayer(self):
        """
        """
        name = self.__getWidget('ng_black_name_entry').get_text()
        if self.__getWidget('ng_black_human_radio').get_active():
            if len(name) == 0:
                return None
            return (name, None)
        
        elif self.__getWidget('ng_black_ai_radio').get_active():
            if len(name) == 0:
                return None
            aiType = self.__getAIType(self.__getWidget('ng_black_ai_type_combo'))
            if aiType is None:
                return None
            return (name, aiType)
        
        elif self.__getWidget('ng_black_network_radio').get_active():
            return (None, None)
  
    def __getLocalPlayer(self):
        """
        """
        name = self.__getWidget('ng_local_name_entry').get_text()
        if len(name) == 0:
            return None

        if self.__getWidget('ng_local_human_radio').get_active():
            return (name, None)
        
        elif self.__getWidget('ng_local_ai_radio').get_active():
            aiType = self.__getAIType(self.__getWidget('ng_local_ai_type_combo'))
            if aiType is None:
                return None
            return (name, aiType)
        
        elif self.__getWidget('ng_local_spectator_radio').get_active():
            return (name, None) # FIXME: Distinguish from human

    def __testReady(self):
        ready = True

        # Depends on what page we start on
        notebook = self.__getWidget('ng_notebook')
        if notebook is None:
            page = 0
        else:
            page = notebook.get_current_page()
        
        # Start game
        if page == 0:
            # Must have a name for the game
            name = self.__getWidget('ng_game_name_entry').get_text()
            if len(name) == 0:
                ready = False
            
            # Must have two valid players
            white = self.__getWhitePlayer()
            black = self.__getBlackPlayer()
            if white is None or black is None:
                ready = False
                
        # Join game
        elif page == 1:
            # Must have a selected server
            if self.__serverList.getSelected() is None:
                ready = False

            # Must have a valid local player
            player = self.__getLocalPlayer()
            if player is None:
                ready = False
                
            # FIXME: Some games do not allow spectators

        else:
            pass #FIXME: raise ABigException

        # Can only click OK if have enough information
        self.__getWidget('ng_ok_button').set_sensitive(ready)

        # Disable properties if toggle buttons not selected
        self.__getWidget('ng_white_ai_properties_table').set_sensitive(self.__getWidget('ng_white_ai_radio').get_active())
        self.__getWidget('ng_black_ai_properties_table').set_sensitive(self.__getWidget('ng_black_ai_radio').get_active())
        table = self.__getWidget('ng_local_ai_properties_table')
        if table is not None:
            table.set_sensitive(self.__getWidget('ng_local_ai_radio').get_active())
        
    def __startGame(self):
        # Depends on what page we start on
        notebook = self.__getWidget('ng_notebook')
        if notebook is None:
            page = 0
        else:
            page = notebook.get_current_page()
        
        # Start game
        if page == 0:        
            # FIXME: Game properties
            gameName = self.__getWidget('ng_game_name_entry').get_text()
            allowSpectators = True

            # Get the players
            white = self.__getWhitePlayer()
            black = self.__getBlackPlayer()
            assert(white is not None)
            assert(black is not None)
            
            # Inform the child class
            self.__mainUI.onGameStart(gameName, allowSpectators, white[0], white[1], black[0], black[1], self.__moves)

        # Join game
        elif page == 1:
            player = self.__getLocalPlayer()
            assert(player is not None)

            # Joining a server
            server = self.__serverList.getSelected()
            assert(server is not None)
            self.__mainUI.onGameJoin(player[0], player[1], server.getHost(), server.getServerId(), server.getGameName())

        else:
            pass #FIXME: raise ABigException
        
    # Gtk+ signal handlers

    def _on_ng_find_servers_button_clicked(self, widget, data=None):
        """Gtk+ callback"""
        host = self.__getWidget('ng_server_entry').get_text()
        if host == '':
            self.__mainUI.onNetworkServerSearch()
        else:
            self.__mainUI.onNetworkServerSearch(host)

    def _on_ng_search_server_entry_changed(self, widget, data=None):
        """Gtk+ callback"""
        # FIXME: Change colour back to default
        pass
    
    def _on_ng_properties_changed(self, widget, *data):
        """Gtk+ callback"""
        self.__testReady()

    def _on_response(self, widget, response_id, data = None):
        """Gtk+ callback"""
        if response_id == gtk.RESPONSE_OK:
            self.__startGame()
        self.__getWidget('new_game_dialog').destroy()

class GtkLoadGameDialog:
    """
    """
    
    __mainUI = None
    __gui = None
    
    def __init__(self, mainUI):
        """
        """
        self.__mainUI = mainUI
        
        # Load the UI
        self.__gui = gtk.glade.XML(UI_DATA_DIR + 'load_game.glade', domain = 'glchess')
        self.__gui.signal_autoconnect(self)
        
        fileChooser = self.__gui.get_widget('filechooserwidget')
        
        # Filter out non PGN files by default
        pgnFilter = gtk.FileFilter()
        pgnFilter.set_name('PGN files')
        pgnFilter.add_pattern('*.pgn')
        fileChooser.add_filter(pgnFilter)
        
        allFilter = gtk.FileFilter()
        allFilter.set_name('All files')
        allFilter.add_pattern('*')
        fileChooser.add_filter(allFilter)
        
        dialog = self.__gui.get_widget('game_load_dialog')
        dialog.show()
        
    def __getFilename(self):
        return self.__gui.get_widget('filechooserwidget').get_filename()
    
    def _on_file_changed(self, widget, data = None):
        """Gtk+ callback"""
        isFile = os.path.isfile(self.__getFilename())
        self.__gui.get_widget('open_button').set_sensitive(isFile)
        self.__gui.get_widget('properties_button').set_sensitive(isFile)
        
    def _on_load_game(self, widget, data = None):
        """Gtk+ callback"""
        self.__mainUI.onGameLoad(self.__getFilename(), False)
        self._on_close(widget, data)
    
    def _on_configure_game(self, widget, data = None):
        """Gtk+ callback"""
        self.__mainUI.onGameLoad(self.__getFilename(), True)
        self._on_close(widget, data)
        
    def _on_close(self, widget, data = None):
        """Gtk+ callback"""
        self.__gui.get_widget('game_load_dialog').destroy()
        
class GtkSaveGameDialog:
    """
    """
    
    # The main UI
    __mainUI = None

    # The view that is being saved
    __view = None
    
    # The GUI
    __gui = None
    
    def __init__(self, mainUI, view):
        """
        """
        self.__mainUI = mainUI
        self.__view = view
        
        # Load the UI
        self.__gui = gtk.glade.XML(UI_DATA_DIR + 'save_game.glade', domain = 'glchess')
        self.__gui.signal_autoconnect(self)
        
        # Filter out non PGN files by default
        dialog = self.__gui.get_widget('dialog')
        pgnFilter = gtk.FileFilter()
        pgnFilter.set_name('PGN files')
        pgnFilter.add_pattern('*.pgn')
        dialog.add_filter(pgnFilter)
        
        allFilter = gtk.FileFilter()
        allFilter.set_name('All files')
        allFilter.add_pattern('*')
        dialog.add_filter(allFilter)
        
    def _on_save(self, widget, data = None):
        """Gtk+ callback"""
        dialog = self.__gui.get_widget('dialog')
        self.__mainUI._saveView(self.__view, dialog.get_filename())
        dialog.destroy()

    def _on_close(self, widget, data = None):
        """Gtk+ callback"""
        dialog = self.__gui.get_widget('dialog')
        dialog.destroy()
        self.__mainUI._saveView(self.__view, None)

class GtkErrorDialog:
    """
    """
    
    __gui = None
    
    def __init__(self, title, contents):
        """
        """
        self.__gui = gtk.glade.XML(UI_DATA_DIR + 'error_dialog.glade', domain = 'glchess')
        self.__gui.signal_autoconnect(self)
        
        self.__gui.get_widget('title_label').set_markup('<b><big>' + title + '</big></b>')
        self.__gui.get_widget('content_label').set_text(contents)
        
        dialog = self.__gui.get_widget('dialog').show_all()

    def _on_close(self, widget, data = None):
        """Gtk+ callback"""
        dialog = self.__gui.get_widget('dialog').destroy()
