__author__ = 'Robert Ancell <bob27@users.sourceforge.net>'
__license__ = 'GNU General Public License Version 2'
__copyright__ = 'Copyright 2005-2006  Robert Ancell'

__all__ = ['GtkView', 'GtkUI']

# Extend base UI classes
import os
import time
import gettext
import ConfigParser

import gobject
import gtk
import gtk.glade
import gtk.gdk
import gtk.gtkgl

import glchess.ui
import dialogs

# TEMP
from OpenGL.GL import *

UI_DATA_DIR = 'lib/glchess/gtkui/'
if not os.path.isdir(UI_DATA_DIR):
    UI_DATA_DIR = '/usr/share/games/glchess/gui/'
    
class GtkViewArea(gtk.DrawingArea, gtk.gtkgl.Widget):
    """Custom widget to render an OpenGL scene"""
    # The view this widget is rendering
    __view = None
    
    # TODO
    __glDrawable = None
    
    def __init__(self, view, glConfig):
        """
        """
        gtk.DrawingArea.__init__(self)
        
        self.__view = view

        # Allow notification of button presses
        self.add_events(gtk.gdk.BUTTON_PRESS_MASK | gtk.gdk.BUTTON_RELEASE_MASK | gtk.gdk.BUTTON_MOTION_MASK)
        
        # Make openGL drawable
        self.set_gl_capability(glConfig)# FIXME:, share_list=glContext)

        # Connect signals
        self.connect('realize', self.__init)
        self.connect('configure_event', self.__configure)
        self.connect('expose_event', self.__expose)
        self.connect('button_press_event', self.__button_press)
        self.connect('button_release_event', self.__button_release)
        
    # Public methods
        
    def redraw(self):
        """Request this widget is redrawn"""
        #FIXME: Check this is valid
        self.window.invalidate_rect(self.allocation, False)
    
    # Private methods
        
    def __startGL(self):
        """Get the OpenGL context"""
        assert(self.__glDrawable is None)
        
        # Obtain a reference to the OpenGL drawable
        # and rendering context.
        glDrawable = self.get_gl_drawable()
        glContext = self.get_gl_context()

        # OpenGL begin.
        if not glDrawable.gl_begin(glContext):
            return
        
        self.__glDrawable = glDrawable
        
    def __endGL(self):
        """Free the OpenGL context"""
        assert(self.__glDrawable is not None)
        self.__glDrawable.gl_end()
        self.__glDrawable = None
        
    def __init(self, widget):
        """Gtk+ signal"""
        self.__startGL()
        print 'Using OpenGL:'
        print 'VENDOR=' + glGetString(GL_VENDOR)
        print 'RENDERER=' + glGetString(GL_RENDERER)
        print 'VERSION=' + glGetString(GL_VERSION)
        print 'EXTENSIONS=' + glGetString(GL_EXTENSIONS)
        self.__view.onReshape(widget.allocation.width, widget.allocation.height)
        self.__endGL()
        
    def __configure(self, widget, event):
        """Gtk+ signal"""
        self.__startGL()
        self.__view.onReshape(event.width, event.height)
        self.__endGL()
    
    def __expose(self, widget, data = None):
        """Gtk+ signal"""
        self.__startGL()
        
        # Get the scene rendered
        self.__view.onRender()
        
        # Paint this
        if self.__glDrawable.is_double_buffered():
            self.__glDrawable.swap_buffers()
        else:
            glFlush()

        self.__endGL()
        
    def __button_press(self, widget, event):
        """Gtk+ signal"""
        #print 'button_press' + str((event.x, event.y))
        self.__startGL()
        self.__view.onSelect(event.x, event.y)
        self.__endGL()
        
    def __button_release(self, widget, event):
        """Gtk+ signal"""
        #print 'button_release' + str((event.x, event.y))
        self.__startGL()
        self.__view.onDeselect(event.x, event.y)
        self.__endGL()
        
class GtkView(glchess.ui.View):
    """
    """
    # The UI this view belongs to
    __ui           = None
    
    __name         = None
    
    # The widget to render the scene to
    _widget        = None # TEMP, make more private
    
    # A Gtk+ tree model to store the move history
    __moveModel    = None
    __selectedMove = -1
    
    def __init__(self, ui, name):
        """Constructor for a view.
        
        'ui' is the UI this view belongs to (glchess.ui.UI).
        'name' is the name of the view (string).
        """
        self.__ui = ui
        self.__name = name
        self._widget = GtkViewArea(self, ui._glConfig)

        # Make a model for navigation
        model = gtk.ListStore(int, str)
        iter = model.append()
        model.set(iter, 0, 0, 1, gettext.gettext('Game Start'))
        self.__moveModel = model

        ui._addView(name, self)
        self._widget.show_all()
    
    # Public methods required by glchess.ui.View
    
    def render(self):
        """Extends glchess.ui.View"""
        self._widget.redraw()
        
    def addMove(self, move):
        """Extends glchess.ui.View"""
        # FIXME: Make a '@ui' player who watches for these itself?
        iter = self.__moveModel.append()
        string = '%2i. ' % ((move.number - 1) / 2 + 1)
        if move.number % 2 == 0:
            string += '... '
        string += move.sanMove
        self.__moveModel.set(iter, 0, move.number, 1, string)
        
        # If is the current view and tracking the game select this
        if self.__selectedMove == -1:
            self.__ui._updateViewButtons()
    
    def close(self):
        """Extends glchess.ui.View"""
        self.__ui._removeView(self)
    
    # Public methods

    def _getModel(self):
        """
        """
        return (self.__moveModel, self.__selectedMove)
        
    def _setMoveNumber(self, moveNumber):
        """Set the move number this view requests.
        
        'moveNumber' is the move number to use (integer).
        """
        self.__selectedMove = moveNumber
        self.onSetMoveNumber(moveNumber)
        
class GtkGameNotebook(gtk.Notebook):
    """
    """
    
    _glConfig       = None
    
    __splashScreen  = None
    
    __viewsByWidget = None
    
    def __init__(self):
        """
        """
        self.__viewsByWidget = {}
        
        gtk.Notebook.__init__(self)
        
        # Make the tabs scrollable so the area is not resized
        self.set_scrollable(True)
        
        # Configure openGL
        display_mode = (gtk.gdkgl.MODE_RGB | gtk.gdkgl.MODE_DEPTH | gtk.gdkgl.MODE_DOUBLE)
        try:
            self._glConfig = gtk.gdkgl.Config(mode=display_mode)
        except gtk.gdkgl.NoMatches:
            display_mode &= ~gtk.gdkgl.MODE_DOUBLE
            self._glConfig = gtk.gdkgl.Config(mode=display_mode)
            
        self.__insertSplashScreen()
        self.set_show_tabs(False)

    def addView(self, name, view):
        """Add a view into the notebook.
        
        'name' is the name of the view.
        'view' is the view containing the callbacks.
        """
        # Get rid of the splashscreen
        if self.__splashScreen is not None:
            self.remove_page(0)
            self.__splashScreen = None
            
        self.__viewsByWidget[view._widget] = view
        
        page = self.append_page(view._widget)
        self.set_tab_label_text(view._widget, name)
        
        self.set_current_page(page)
        
        self.__updateTabVisibleState()
    
    def getView(self, pageNumber = None):
        """Get the view at a given page number.
        
        'pageNumber' is the page to check or None to get the selected page.
        
        Return the view (GtkView) on this page or None if no view here.
        """
        if pageNumber is None:
            # If splashscreen present then there is no view
            if self.__splashScreen is None:
                widget = self.get_nth_page(self.get_current_page())
            else:
                return None
        else:
            widget = self.get_nth_page(pageNumber)
            
        return self.__viewsByWidget[widget]
    
    def removeView(self, view):
        """Remove a view from the notebook.
        
        'view' is the view to remove.
        """
        self.remove_page(self.page_num(view._widget))
        self.__updateTabVisibleState()
        if self.get_n_pages() == 0:
            self.__insertSplashScreen()
        self.__viewsByWidget.pop(view._widget)

    def __insertSplashScreen(self):
        ui = gtk.glade.XML(UI_DATA_DIR + 'splashscreen.glade', 'about_box', domain = 'glchess')
        self.__splashScreen = ui.get_widget('about_box')
        self.__splashScreen.show()
        self.append_page(self.__splashScreen, gtk.Label('!'))
        
    def __updateTabVisibleState(self):
        """
        """
        # Only show tabs if there is more than one game
        self.set_show_tabs(self.get_n_pages() > 1)

class GtkUI(glchess.ui.UI):
    """
    """

    # The Gtk+ GUI
    __gui             = None
    
    # The time stored for animation
    __lastTime        = None

    # The notebook containing games
    __notebook        = None
    
    # The Gtk+ list model of the available AIs
    __aiModel         = None
    
    # Dictionary of save game dialogs keyed by view
    __saveGameDialogs = None
    
    # Dictionary of configuration options
    __config          = None
    __applyingConfig  = False
    
    def __init__(self):
        """Constructor for a GTK+ glChess GUI"""
        self.__saveGameDialogs = {}
        
        self.__gui = gtk.glade.XML(UI_DATA_DIR + 'glchess.glade', domain = 'glchess')
        self.__gui.signal_autoconnect(self)

        # Make a notebook for the games
        self.__notebook = GtkGameNotebook()
        self._glConfig = self.__notebook._glConfig # TEMP
        self.__notebook.connect_after('switch-page', self._on_view_changed)
        self.__getWidget('game_viewport').add(self.__notebook)
        self.__notebook.show_all()
        
        # Create the model for the AI types
        self.__aiModel = gtk.ListStore(str)
        
        combo = self.__getWidget('history_combo')
        cell = gtk.CellRendererText()
        combo.pack_start(cell, False)
        combo.add_attribute(cell, 'text', 1)
        
        # Load UI preferences
        self.__config = {}
        self.__loadConfig()

    # Public methods
    
    def addTimer(self, period, callback):
        """Add a timeout .
        
        'period' is the period for the timer in milliseconds (integer).
        'callback' is the method/function to call when the timer expires.
        
        The timer will be continuously called while 'callback' returns True.
        """
        gobject.timeout_add(period, callback)

    def addAIEngine(self, name):
        """Register an AI engine.
        
        'name' is the name of the engine.
        TODO: difficulty etc etc
        """
        iter = self.__aiModel.append()
        self.__aiModel.set(iter, 0, name)

    def reportError(self, title, error):
        """Extends glchess.ui.UI"""
        dialogs.GtkErrorDialog(title, error)
        
    def reportGameLoaded(self, gameName = None,
                         whiteName = None, blackName = None,
                         whiteAI = None, blackAI = None, moves = None):
        """Extends glchess.ui.UI"""
        dialogs.GtkNewGameDialog(self, self.__aiModel, gameName = gameName,
                                 whiteName = whiteName, whiteAI = whiteAI,
                                 blackName = blackName, blackAI = blackAI, moves = moves)

    def run(self):
        """Run the UI.
        
        This method will not return.
        """
        # Make an animation timer
        self.__lastTime = time.time()
        gobject.timeout_add(10, self.__animate)
        
        gtk.main()

    # Protected methods

    def _addView(self, name, view):
        """Add a view into the UI.
        
        'name' is the name of the view.
        'view' is the view to add (extends ui.View).
        """
        self.__notebook.addView(name, view)
    
    def _saveView(self, view, path):
        """
        """
        self.__saveGameDialogs.pop(view)
        if path is not None:
            self.onGameSave(view, path)
            
    def _removeView(self, view):
        """Remove a view from the UI.
        
        'view' is the view to remove.
        """
        self.__notebook.removeView(view)
        self._updateViewButtons()
            
    # Private methods
    
    def __loadConfig(self):
        """
        """
        name = os.path.expanduser('~/.glchess/gtkui.ini')
        cp = ConfigParser.ConfigParser()
        cp.read(name)
        
        try:
            items = cp.items('ui')
        except ConfigParser.NoSectionError:
            # If file is missing set defaults
            self.__config['show_toolbar'] = 'True'
            self.__config['show_history'] = 'True'
            self.__saveConfig()
        else:
            for (name, value) in items:
                self.__config[name] = value
            
        self.__applyConfig()
        
    def __setConfig(self, name, value):
        """
        """
        if self.__applyingConfig:
            return
        
        self.__config[name] = value
        self.__applyConfig()
        self.__saveConfig()
        
    def __saveConfig(self):
        """
        """
        name = os.path.expanduser('~/.glchess/gtkui.ini')
        try:
            f = file(name, 'w')
        except IOError, e:
            print 'Unable to save config: ' + str(e)
            return
        
        cp = ConfigParser.ConfigParser()
        cp.add_section('ui')
        for (name, value) in self.__config.iteritems():
            cp.set('ui', name, value)
        cp.write(f)
        
    def __applyConfig(self):
        """
        """
        # Stop recursion
        self.__applyingConfig = True
        
        # Show/hide the toolbar
        toolbar = self.__getWidget('toolbar')
        menu = self.__getWidget('menu_view_toolbar')
        if self.__config['show_toolbar'] == 'True':
            menu.set_active(True)
            toolbar.show()
        else:
            menu.set_active(False)
            toolbar.hide()
            
        # Show/hide the history
        box = self.__getWidget('navigation_box')
        menu = self.__getWidget('menu_view_history')
        if self.__config['show_history'] == 'True':
            menu.set_active(True)
            box.show()
        else:
            menu.set_active(False)
            box.hide()
            
        self.__applyingConfig = False
    
    def __animate(self):
        # Get the timestep, if it is less than zero or more than a second
        # then the system clock was probably changed.
        now = time.time()
        step = now - self.__lastTime
        if step < 0.0:
            step = 0.0
        elif step > 1.0:
            step = 1.0
        self.__lastTime = now
        
        # Animate!
        self.onAnimate(step)

        # Keep the timer running (deletes itself when returns False)
        return True

    def __getWidget(self, name):
        return self.__gui.get_widget(name)
    
    def _on_show_toolbar_clicked(self, widget, data = None):
        """Gtk+ callback"""
        if widget.get_active():
            value = 'True'
        else:
            value = 'False'
        self.__setConfig('show_toolbar', value)

    def _on_show_history_clicked(self, widget, data = None):
        """Gtk+ callback"""
        if widget.get_active():
            value = 'True'
        else:
            value = 'False'
        self.__setConfig('show_history', value)
            
    def _on_history_combo_changed(self, widget, data = None):
        """Gtk+ callback"""
        model = widget.get_model()
        iter = widget.get_active_iter()
        if iter is None:
            return
        
        # Get the move number
        moveNumber = model.get_value(iter, 0)
        
        string = 'Show move number: ' + str(moveNumber)
        if moveNumber == len(model) - 1:
            string += ' (latest)'
            moveNumber = -1
        
        view = self.__notebook.getView()
        if view is not None:
            view._setMoveNumber(moveNumber)
        
    def __selectMoveNumber(self, moveNumber):
        """FIXME
        """
        combo = self.__getWidget('history_combo')
        
        # Limit moves to the maximum value
        maxNumber = len(combo.get_model())

        # Allow negative indexing
        if moveNumber < 0:
            moveNumber = maxNumber + moveNumber
        if moveNumber < 0:
            moveNumber = 0
        if moveNumber >= maxNumber:
            moveNumber = maxNumber - 1
        
        combo.set_active(moveNumber)

    def __selectMoveNumberRelative(self, offset):
        """FIXME
        """
        combo = self.__getWidget('history_combo')
        selected = combo.get_active()
        maxNumber = len(combo.get_model())
        new = selected + offset
        if new < 0:
            new = 0
        elif new >= maxNumber:
            new = maxNumber - 1
        self.__selectMoveNumber(new)

    def _on_history_start_clicked(self, widget, data = None):
        """Gtk+ callback"""
        self.__selectMoveNumber(0)

    def _on_history_previous_clicked(self, widget, data = None):
        """Gtk+ callback"""
        self.__selectMoveNumberRelative(-1)

    def _on_history_next_clicked(self, widget, data = None):
        """Gtk+ callback"""
        self.__selectMoveNumberRelative(1)

    def _on_history_latest_clicked(self, widget, data = None):
        """Gtk+ callback"""
        self.__selectMoveNumber(-1)

    def _on_view_changed(self, widget, page, pageNum, data = None):
        """Gtk+ callback"""
        self._updateViewButtons()
    
    def _updateViewButtons(self):
        """
        """
        view = self.__notebook.getView()
        self.__getWidget('end_game_button').set_sensitive(view is not None)
        self.__getWidget('save_game_button').set_sensitive(view is not None)
        self.__getWidget('menu_save_item').set_sensitive(view is not None)
        self.__getWidget('menu_end_game_item').set_sensitive(view is not None)

        combo = self.__getWidget('history_combo')
        if view is None:
            combo.set_model(None)
        else:
            (model, selected) = view._getModel()
            combo.set_model(model)
            if selected < 0:
                selected = len(model) + selected
            combo.set_active(selected)
        self.__getWidget('navigation_box').set_sensitive(view is not None)

    def _on_new_game_button_clicked(self, widget, data = None):
        """Gtk+ callback"""
        dialogs.GtkNewGameDialog(self, self.__aiModel)
        
    def _on_open_game_button_clicked(self, widget, data = None):
        """Gtk+ callback"""
        dialogs.GtkLoadGameDialog(self)
        
    def _on_save_game_button_clicked(self, widget, data = None):
        """Gtk+ callback"""
        view = self.__notebook.getView()
        if not self.__saveGameDialogs.has_key(view):
            self.__saveGameDialogs[view] = dialogs.GtkSaveGameDialog(self, view)
        
    def _on_end_game_button_clicked(self, widget, data = None):
        """Gtk+ callback"""
        view = self.__notebook.getView()
        assert(view is not None)
        view.onClose()

    def _on_quit(self, widget, data = None):
        """Gtk+ callback"""
        self.onQuit()

if __name__ == '__main__':
    ui = GtkUI()
    ui.run()
